/* -*-c-*-
 *
 * $Id: pkcs1.c,v 1.3 2000/10/08 12:07:04 mdw Exp $
 *
 * PKCS#1 1.5 packing
 *
 * (c) 2000 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------* 
 *
 * This file is part of Catacomb.
 *
 * Catacomb is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * Catacomb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with Catacomb; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

/*----- Revision history --------------------------------------------------* 
 *
 * $Log: pkcs1.c,v $
 * Revision 1.3  2000/10/08 12:07:04  mdw
 * Don't do arithmetic on @void *@ pointers.
 *
 * Revision 1.2  2000/07/05 17:49:48  mdw
 * Fix decoding functions, so that they don't run off the end of the
 * buffer.
 *
 * Revision 1.1  2000/07/01 11:17:38  mdw
 * New support for PKCS#1 message encoding.
 *
 */

/*----- Header files ------------------------------------------------------*/

#include <string.h>

#include <mLib/bits.h>
#include <mLib/dstr.h>

#include "grand.h"
#include "pkcs1.h"

/*----- Main code ---------------------------------------------------------*/

/* --- @pkcs1_cryptencode@ --- *
 *
 * Arguments:	@const void *msg@ = pointer to message data
 *		@size_t msz@ = size of message data
 *		@void *buf@ = pointer to output buffer
 *		@size_t sz@ = size of the output buffer
 *		@void *p@ = pointer to PKCS1 parameter block
 *
 * Returns:	Zero if all went well, negative on failure.
 *
 * Use:		Implements the operation @EME-PKCS1-V1_5-ENCODE@, as defined
 *		in PKCS#1 v. 2.0 (RFC2437).
 */

int pkcs1_cryptencode(const void *msg, size_t msz, void *buf, size_t sz,
		      void *p)
{
  pkcs1 *pp = p;
  grand *r = pp->r;
  octet *q, *qq;
  size_t i, n;

  /* --- Ensure that the buffer is sensibly sized --- */

  if (pp->epsz + msz + 11 > sz)
    return (-1);

  /* --- Fill in the buffer --- */

  q = buf;
  qq = q + sz;
  *q++ = 0;
  *q++ = 2;
  n = sz - msz - pp->epsz - 3;
  r->ops->fill(r, q, n);
  for (i = 0; i < n; i++) {
    if (*q == 0)
      *q = r->ops->range(r, 255) + 1;
    q++;
  }
  *q++ = 0;
  memcpy(q, pp->ep, pp->epsz);
  q += pp->epsz;
  memcpy(q, msg, msz);
  return (0);
}

/* --- @pkcs1_cryptdecode@ --- *
 *
 * Arguments:	@const void *buf@ = pointer to encoded buffer
 *		@size_t sz@ = size of the encoded buffer
 *		@dstr *d@ = pointer to destination string
 *		@void *p@ = pointer to PKCS1 parameter block
 *
 * Returns:	The length of the output string if successful, negative on
 *		failure.
 *
 * Use:		Implements the operation @EME-PKCS1-V1_5-DECODE@, as defined
 *		in PKCS#1 v. 2.0 (RFC2437).
 */

int pkcs1_cryptdecode(const void *buf, size_t sz, dstr *d, void *p)
{
  pkcs1 *pp = p;
  const octet *q, *qq;
  size_t n, i;

  /* --- Check the size of the block looks sane --- */

  if (pp->epsz + 11 > sz)
    return (-1);
  q = buf;
  qq = q + sz;

  /* --- Ensure that the block looks OK --- */

  if (*q++ != 0 || *q++ != 2)
    return (-1);

  /* --- Check the nonzero padding --- */

  i = 0;
  while (*q != 0 && q < qq)
    i++, q++;
  if (i < 8 || qq - q < pp->epsz + 1)
    return (-1);
  q++;

  /* --- Check the encoding parameters --- */

  if (memcmp(q, pp->ep, pp->epsz) != 0)
    return (-1);
  q += pp->epsz;

  /* --- Done --- */

  n = qq - q;
  dstr_putm(d, q, n);
  return (n);
}

/* --- @pkcs1_sigencode@ --- *
 *
 * Arguments:	@const void *msg@ = pointer to message data
 *		@size_t msz@ = size of message data
 *		@void *buf@ = pointer to output buffer
 *		@size_t sz@ = size of the output buffer
 *		@void *p@ = pointer to PKCS1 parameter block
 *
 * Returns:	Zero if all went well, negative on failure.
 *
 * Use:		Implements the operation @EMSA-PKCS1-V1_5-ENCODE@, as defined
 *		in PKCS#1 v. 2.0 (RFC2437).
 */

int pkcs1_sigencode(const void *msg, size_t msz, void *buf, size_t sz,
		      void *p)
{
  pkcs1 *pp = p;
  octet *q, *qq;
  size_t n;

  /* --- Ensure that the buffer is sensibly sized --- */

  if (pp->epsz + msz + 11 > sz)
    return (-1);

  /* --- Fill in the buffer --- */

  q = buf;
  qq = q + sz;
  *q++ = 0;
  *q++ = 1;
  n = sz - msz - pp->epsz - 3;
  memset(q, 0xff, n);
  q += n;
  *q++ = 0;
  memcpy(q, pp->ep, pp->epsz);
  q += pp->epsz;
  memcpy(q, msg, msz);
  return (0);
}

/* --- @pkcs1_sigdecode@ --- *
 *
 * Arguments:	@const void *buf@ = pointer to encoded buffer
 *		@size_t sz@ = size of the encoded buffer
 *		@dstr *d@ = pointer to destination string
 *		@void *p@ = pointer to PKCS1 parameter block
 *
 * Returns:	The length of the output string if successful, negative on
 *		failure.
 *
 * Use:		Implements the operation @EMSA-PKCS1-V1_5-DECODE@, as defined
 *		in PKCS#1 v. 2.0 (RFC2437).
 */

int pkcs1_sigdecode(const void *buf, size_t sz, dstr *d, void *p)
{
  pkcs1 *pp = p;
  const octet *q, *qq;
  size_t i, n;

  /* --- Check the size of the block looks sane --- */

  if (pp->epsz + 10 > sz)
    return (-1);
  q = buf;
  qq = q + sz;

  /* --- Ensure that the block looks OK --- */

  if (*q++ != 0 || *q++ != 1)
    return (-1);

  /* --- Check the padding --- */

  i = 0;
  while (*q == 0xff && q < qq)
    i++, q++;
  if (i < 8 || qq - q < pp->epsz + 1 || *q++ != 0)
    return (-1);

  /* --- Check the encoding parameters --- */

  if (memcmp(q, pp->ep, pp->epsz) != 0)
    return (-1);
  q += pp->epsz;

  /* --- Done --- */

  n = qq - q;
  dstr_putm(d, q, n);
  return (n);
}

/*----- That's all, folks -------------------------------------------------*/
